# -*- coding: utf-8 -*-

"""Pashword package.

This Python package allows you to easily and securely manage your
passwords.

"""

__author__ = 'Dunstan Becht'

from configparser import ConfigParser
from logging import basicConfig
from pkgutil import get_data


config = ConfigParser()
config.read_string(get_data(__package__, 'default.conf').decode())

basicConfig(level=config.get(__name__, 'log'))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""Test script.

This script executes the unit tests for the package.

"""

from contextlib import contextmanager
from json import load, dumps
from pathlib import Path
from shutil import copytree
from subprocess import CalledProcessError, run
from sys import stderr
from tempfile import TemporaryDirectory, TemporaryFile
from unittest import main, TestCase

from pashword.core import pash, sets, book


dir_tests = Path(__file__).parent
dir_material = dir_tests/'material'

with open(dir_tests/'references.json', 'r', encoding='utf-8') as ref_file:
    ref = load(ref_file)


def runcheck(command):
    """Run a command and check the exit code.

    If the exit code of the executed command is different from 0, a
    RuntimeError containing the output of the command is raised.

    Parameters
    ----------
    command : str
        Command to be runned.

    Raises
    ------
    RuntimeError
        If the command exited with an error code.

    """
    with TemporaryFile() as file:
        try:
            run(command, shell=True, check=True, stdout=file, stderr=file)
        except CalledProcessError as error:
            file.seek(0)
            raise RuntimeError(file.read().decode()) from error


@contextmanager
def temporary(directory='.'):
    """Return a temporary clone of the target test directory.

    Parameters
    ----------
    directory : str
        Name of the directory to clone from material.

    Returns
    -------
    str
        Path to the cloned target directory.

    """
    temp_dir = TemporaryDirectory()
    path = Path(temp_dir.name)
    copytree(dir_material/directory, path, dirs_exist_ok=True)
    try:
        yield path
    finally:
        temp_dir.cleanup()


def display(data):
    """Print data in JSON format.

    Parameters
    ----------
    data : dict or list
        Data to be displayed.

    """
    print("\n"+dumps(data, indent=4), file=stderr)


class TestCore(TestCase):
    """Class for testing core subpackage."""

    def test_metacharacters(self):
        """Test metacharacter set retrieval."""
        for metacharacter, charaters in ref['sets'].items():
            self.assertEqual(charaters, sets.get(metacharacter))

    def test_combinations(self):
        """Test combination number."""
        for form, number in ref['combinations'].items():
            self.assertEqual(number, sets.combinations(form))

    def test_hexdigest(self):
        """Test hash value."""
        for key, salt, value in ref['hexdigest']:
            self.assertEqual(value, pash.hexdigest(key, salt))

    def test_password(self):
        """Test pashword."""
        for key, salt, form, value in ref['password']:
            self.assertEqual(value, pash.password(key, salt, form))

    def test_same(self):
        """Test key match."""
        with temporary('logins') as material:
            key = 'key'
            path = material/'validate.json'
            pash.save(key, path)
            self.assertTrue(pash.same(key, path))

    def test_load(self):
        """Test password book loading."""
        with temporary('logins') as material:
            data = book.load(material/'user.conf')
        self.assertTrue(ref['loading'], data)

    def test_matching(self):
        """Test name matching."""
        data = book.matching(ref['matching']['input'], '*1')
        self.assertEqual(ref['matching']['output'], data)

    def test_decode(self):
        """Test decoding."""
        with temporary('logins') as material:
            data = book.load(material/'user.conf')
            decoded = book.decode(data, ref['decoding']['key'])
            self.assertEqual(ref['decoding']['output'], decoded)


class TestCommandLineInterface(TestCase):
    """Class for testing the CLI."""

    def test_help(self):
        """Test help option."""
        runcheck("pashword --help")
        runcheck("pashword read --help")
        runcheck("pashword sort --help")

    def test_features(self):
        """Test all features."""
        with temporary('logins') as material:
            conf = material/'user.conf'
            json = material/'user.json'
            prompt = dir_material/'inputs'/'read.txt'
            runcheck(f"pashword read {conf} < {prompt}")
            runcheck(f"pashword read {conf} --hash {json} < {prompt}")
            runcheck(f"pashword read {conf} --hash {json} < {prompt}")
            runcheck(f"pashword sort {conf} < {prompt}")
            runcheck(f"pashword read {conf} < {prompt}")


if __name__ == '__main__':
    main(buffer=True)



# -*- coding: utf-8 -*-

"""Hashing and passwords.

This module provides the hash features for password creation and
secret key validation.

"""

from hashlib import new
from json import dump, load
from logging import getLogger
from uuid import uuid4

from pashword import config
from pashword.core.sets import get


logger = getLogger(__name__)


def digest(key, salt):
    """Return the hash value in binary.

    Parameters
    ----------
    key : str
        Secret key.
    salt : str
        Hash salt.

    Returns
    -------
    bytes
        Digest value.

    """
    algo = config.get(__name__, 'hash_algo')
    logger.debug("hashing with '%s'", algo)
    hashing = new(algo, usedforsecurity=True)
    hashing.update(key.encode('utf-8'))
    hashing.update(salt.encode('utf-8'))
    return hashing.digest()


def hexdigest(key, salt):
    """Return the hash value in hexadecimal.

    Parameters
    ----------
    key : str
        Secret key.
    salt : str
        Hash salt.

    Returns
    -------
    str
        Hexadecimal digest value.

    """
    return digest(key, salt).hex()


def password(key, salt, form):
    """Return the account password.

    Parameters
    ----------
    key : str
        Secret key.
    salt : str
        Hash salt.
    form : str
        Password format.

    Returns
    -------
    str
        Password value.

    """
    logger.debug("getting integer hash value")
    series = digest(key, salt)
    index = int.from_bytes(series, byteorder='big', signed=False)
    logger.debug("getting resulting string")
    word = []
    for metacharacter in form:
        characters = get(metacharacter)
        size = len(characters)
        word.append(characters[index % size])
        index //= size
    word = ''.join(word)
    return word


def save(key, path):
    """Save a hash of the key in a file.

    Parameters
    ----------
    key : str
        Secret key.
    path : str
        Path of the file containing the hash.

    """
    salt = str(uuid4())
    data = {
        'salt': salt,
        'hash': hexdigest(key, salt),
    }
    logger.debug("saving hash value in '%s'", path)
    with open(path, 'w', encoding='utf-8') as file:
        dump(data, file)


def same(key, path):
    """Return wether the key is the same as the one from the file.

    Parameters
    ----------
    key : str
        Secret key.
    path : str
        Path of the file containing the hash.

    """
    logger.debug("loading hash value from '%s'", path)
    with open(path, 'r', encoding='utf-8') as file:
        data = load(file)
    return data['hash'] == hexdigest(key, data['salt'])



# -*- coding: utf-8 -*-

"""Password reading feature.

This module implements the functionality that allows to search and
decode a password.

"""

from getpass import getpass
from logging import getLogger
from os.path import isfile
from pathlib import Path
from platform import system

from pashword import config
from pashword.core.book import decode, matching, load
from pashword.core.colors import colorize
from pashword.core.pash import same, save
from pashword.core.sets import combinations

if system() == 'Linux':
    import readline
    _ = readline


logger = getLogger(__name__)


def setup(parser):
    """Configure the parser for the module.

    Parameters
    ----------
    parser : ArgumentParser
        Parser dedicated to the module.

    """
    logger.debug("defining command-line arguments")
    parser.set_defaults(
        func=main,
    )
    parser.add_argument(
        'book',
        help="path to the password book file",
        type=str,
        metavar='path',
    )
    parser.add_argument(
        '--hash',
        help="path to the hash of the key",
        type=str,
        metavar='path',
    )
    exclusive = parser.add_mutually_exclusive_group()
    exclusive.add_argument(
        '--hide',
        action='store_true',
        help="to hide passwords",
    )
    exclusive.add_argument(
        '--show',
        dest='hide',
        action='store_false',
        help="to show passwords",
    )


def main(book, **kwargs):
    """Decode and display the passwords contained in the book.

    Parameters
    ----------
    book : str
        Path to the password book file.

    Keyword Arguments
    -----------------
    hash : str
        Path to the hash of the key.
    hide : bool
        To hide passwords.

    """
    logger.debug("retrieving parameters")
    kwargs.setdefault('hash', config.get(__name__, 'hash'))
    kwargs.setdefault('hide', config.getboolean(__name__, 'hide'))
    colors = {}
    colors['name'] = config.get(__name__, 'color_name')
    colors['show'] = config.get(__name__, 'color_show')
    colors['hide'] = config.get(__name__, 'color_hide')
    colors['warn'] = config.get(__name__, 'color_warn')
    colors['pash'] = colors['hide'] if kwargs['hide'] else colors['show']

    logger.debug("loading password book '%s'", book)
    data = load(book)
    file = colorize(Path(book).name, colors['warn'])
    print(f"reading {file} ({len(data)} sections)")

    logger.debug("filtering sections")
    pattern = input("\nmatching pattern (default: *): ") or '*'
    while pattern:
        filtered = matching(data, pattern)
        print(f"{len(filtered)} matching section(s) found")
        for name in filtered:
            print(f"- {colorize(name, colors['name'])}")
        pattern = input("\nnew pattern or press enter to continue: ")
    if len(filtered) == 0:
        return

    logger.debug("decoding passwords")
    key = get_key(kwargs, colors)
    decoded = decode(filtered, key)
    for name, entries in decoded.items():
        logger.debug("displaying section '%s'", name)
        print(f"\n[{colorize(name, colors['name'])}]")
        for field, value in entries.items():
            if field not in ('password', 'form'):
                print(f"{field} = {value}")
        number = combinations(entries['form'])
        pash = f"{colorize(entries['password'], colors['pash'])}"
        print(f"pash = {pash} ({number:1.0e})")
    print()


def get_key(kwargs, colors):
    """Return the secret key entered by the user.

    Parameters
    ----------
    kwargs : dict
        Keyword arguments of the main function.
    colors : dict
        Color codes.

    Returns
    -------
    str
        The secret key.

    """
    logger.debug("retrieving secret key")
    while True:
        if kwargs['hide']:
            key = getpass("secret key: ")
        else:
            key = input("secret key: ")
        if isfile(kwargs['hash']):
            if same(key, kwargs['hash']):
                return key
            print(colorize("incorrect secret key", colors['warn']))
            logger.debug("incorrect secret key")
        else:
            logger.info("saving key hash value in '%s'", kwargs['hash'])
            save(key, kwargs['hash'])
            print(f"key hash saved in {kwargs['hash']}")
            return key



# -*- coding: utf-8 -*-

"""Character sets.

This module provides the character sets used for the construction of
passwords.

"""

from logging import getLogger

from pashword import config


logger = getLogger(__name__)


def get(metacharacter):
    """Return the set of characters associated to the metacharacter.

    Parameters
    ----------
    metacharacter : str
        Character associated with a set of characters.

    Returns
    -------
    str
        The set of characters associated to the metacharacter.

    """
    logger.debug("retrieving '%s' character set", metacharacter)
    return config.get(__name__, metacharacter, fallback=metacharacter)


def combinations(form):
    """Return the number of possible combinations for the given format.

    Parameters
    ----------
    form : str
        Metacharacters defining the password format.

    Returns
    -------
    int
        The number of possible combinations for the given format.

    """
    logger.debug("computing combinations for '%s'", form)
    number = 1
    for metacharacter in form:
        number *= len(get(metacharacter))
    return number



# -*- coding: utf-8 -*-

"""Pashword command-line interface.

This module allows the user to launch the main features of the package
from a command-line interface.

"""

from argparse import ArgumentParser
from logging import getLogger
from sys import exit as sys_exit

from pashword import config
from pashword._version import version
from pashword.main import read, sort


logger = getLogger(__name__)


def main():
    """Run the command-line interpreter.

    """
    logger.debug("defining command-line arguments")
    parser = ArgumentParser(
        prog=__package__,
        description=__doc__,
    )
    parser.add_argument(
        '-v',
        '--version',
        action='version',
        version=f'%(prog)s {version}',
    )
    parser.add_argument(
        '-c',
        '--config',
        help="custom configuration file path",
        metavar='path',
    )
    parser.add_argument(
        '-l',
        '--log',
        help="logging level",
        choices=('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'),
    )

    logger.debug("defining subparsers")
    subparsers = parser.add_subparsers(
        required=True,
        help="name of the command to run",
    )
    for module in (read, sort):
        subparser = subparsers.add_parser(
            name=module.__name__.rsplit('.', maxsplit=1)[-1],
            help=f"run {module.__name__} module",
        )
        module.setup(subparser)

    logger.debug("parsing command-line arguments")
    args = parser.parse_args()
    kwargs = {key: val for key, val in vars(args).items() if val is not None}
    kwargs.pop('func')
    logger.debug("parsed arguments: %s", kwargs)

    logger.debug("running command")
    try:
        if args.log:
            getLogger().setLevel(args.log)
        if args.config:
            logger.info("loading custom configuration file")
            with open(args.config, encoding='utf-8') as file:
                config.read_file(file)
        logger.debug("executing command function")
        args.func(**kwargs)
    except SystemExit:
        logger.info("program exited")
        sys_exit(0)
    except KeyboardInterrupt:
        logger.info("program interrupted by Control-C")
        sys_exit(130)
    except FileNotFoundError as exception:
        logger.error(exception)
        sys_exit(1)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Configuration file for the Sphinx documentation builder.

For the full list of built-in configuration values, see:
https://www.sphinx-doc.org/en/master/usage/configuration.html

"""

from pashword._version import __version__

# Project information

project = 'pashword'
author = 'Dunstan Becht'
version = __version__
release = __version__

# General configuration

extensions = [
    'sphinx.ext.napoleon',
    'sphinx_rtd_theme',
    'myst_parser',
]

templates_path = []
exclude_patterns = []

# Options for HTML output

# https://sphinx-rtd-theme.readthedocs.io/en/stable/configuring.html
html_theme = 'sphinx_rtd_theme'
html_theme_options = {
    # Table of contents options
    'collapse_navigation': True,
    'sticky_navigation': True,
    'navigation_depth': -1,
    'includehidden': True,
    'titles_only': True,
    # Miscellaneous options
    'display_version': True,
    'prev_next_buttons_location': 'bottom',
    'style_external_links': True,
    'style_nav_header_background': '#d4aa00ff',
}
html_static_path = []
html_show_copyright = False
html_show_sphinx = False



# -*- coding: utf-8 -*-

"""Text colorization.

This module implements features related to text colorization.

"""

from codecs import decode
from logging import getLogger

from pashword import config


logger = getLogger(__name__)


def get_code(name):
    """Return the color code.

    Parameters
    ----------
    name : str
        Name of the color code in the configuration file.

    Returns
    -------
    str
        Color code.

    """
    logger.debug("retrieving '%s' color code", name)
    color = config.get(__name__, name)
    return decode(color, 'unicode_escape')


def colorize(string, color):
    """Return the colorized string.

    Parameters
    ----------
    string : str
        The text to colorize.
    color : str
        The name of the color.

    Returns
    -------
    str
        Colorized text.

    """
    logger.debug("colorizing text with color '%s'", color)
    start = get_code(color)
    end = get_code('end')
    return start + string + end



# -*- coding: utf-8 -*-

"""Password book sorting feature.

This module implements the functionality that allows to sort a
password configuration file.

"""

from logging import getLogger

from pashword.core.book import load, save


logger = getLogger(__name__)


def setup(parser):
    """Configure the parser for the module.

    Parameters
    ----------
    parser : ArgumentParser
        Parser dedicated to the module.

    """
    logger.debug("defining command-line arguments")
    parser.set_defaults(
        func=main,
    )
    parser.add_argument(
        'book',
        help="path to the password book file",
        type=str,
        metavar='path',
    )


def main(book):
    """Sort a password book.

    Parameters
    ----------
    book : str
        Path to the password book file.

    """
    logger.debug("loading password book '%s'", book)
    data = load(book)

    logger.debug("sorting password book '%s'", book)
    ordered = {}
    for name in sorted(data.keys()):
        ordered[name] = data[name]

    logger.debug("saving password book '%s'", book)
    save(book, ordered)



# -*- coding: utf-8 -*-

"""Password book handling.

This module implements functionalities to manipulate password books.

"""

from configparser import ConfigParser
from fnmatch import fnmatch
from logging import getLogger

from pashword.core.pash import password


logger = getLogger(__name__)


def load(path):
    """Return the data contained in the file.

    Parameters
    ----------
    path : str
        Path to the password book.

    Returns
    -------
    dict
        Password book data.

    """
    config = ConfigParser()
    logger.debug("loading book '%s'", path)
    with open(path, 'r', encoding='utf-8') as file:
        config.read_file(file)
    book = {}
    for name in config.sections():
        book[name] = dict(config[name])
    return book


def save(path, book):
    """Save the book into a file.

    Parameters
    ----------
    path : str
        Path of the password book to be saved.
    book : dict
        Password book data to be saved.

    """
    config = ConfigParser()
    config.read_dict(book)
    logger.debug("saving book in '%s'", path)
    with open(path, 'w', encoding='utf-8') as file:
        config.write(file)


def matching(book, pattern):
    """Return the accounts whose name matches the pattern.

    Parameters
    ----------
    book : dict
        Password book.
    pattern : str
        Filtering pattern.

    Returns
    -------
    dict
        Filtered password book.

    """
    logger.debug("filtering with '%s' pattern", pattern)
    filtered = {}
    for name, account in book.items():
        if fnmatch(name, pattern):
            filtered[name] = account
    return filtered


def decode(book, key):
    """Return the book augmented with passwords.

    Parameters
    ----------
    book : dict
        Password book.
    key : str
        Secret key.

    Returns
    -------
    dict
        Decoded password book.

    """
    decoded = {}
    for name, account in book.items():
        logger.debug("decoding section '%s'", name)
        data = account.copy()
        form = data.pop('form')
        salt = ''.join(sorted(data.values())) + name
        word = password(key, salt, form)
        data = {**account, 'password': word}
        decoded[name] = data
    return decoded
